(*Author: Selwyn Hollis, shollis@appliedsymbols.com *)
(* Package Version 1.1, Nov 2004 *)

BeginPackage["RNDS`RestartingNDSolve`", {"Utilities`FilterOptions`"}]; 
Unprotect["RNDS`RestartingNDSolve`*"]; 
ClearAll @@ Names["RNDS`RestartingNDSolve`*"]; 
ClearAll @@ Names["RNDS`RestartingNDSolve`Private`*"];

RestartingNDSolve::usage = "RestartCondition[sys, vars, {t,\!\(t\_0\),\!\(t\_max\)},
{restartcond, restartfunc}, MinRecoveryTime \[Rule] \[Tau]]"; 
MinRecoveryTime::usage = "MinRecoveryTime is an option for RestartingNDSolve.
Its value is the minimum time between restarts. Its default is 1/10"; 
Reporting::usage = "Reporting is an option for RestartingNDSolve that controls
whether restart times are printed as they occur during the computation. Its default is False."; 
TimePrecisionGoal::usage = "TimePrecisionGoal is an option for RestartingNDSolve that specifies the
effective digits of precision sought for restart times. Its default is 8."; 

Options[RestartingNDSolve] = Union[DeleteCases[Options[NDSolve], StoppingTest -> _], 
    {MinRecoveryTime -> 1/10, TimePrecisionGoal -> 8, Reporting -> False}]; 
SetOptions[RestartingNDSolve, MaxSteps -> 10^6, MaxStepSize -> 1]; 

Begin["Private`"]; 

RestartingNDSolve::badf = "Invalid restart function."; 
RestartingNDSolve::badcond = "Invalid restart condition."; 
RestartingNDSolve::badinit = "Initial time t0 must be the same as tmin."; 
RestartingNDSolve::minrec = "Warning: Subinterval length is close to MinRecoveryTime."; 

RestartingNDSolve[sys:{_Equal..}, vars:{_[t_]..}, {t_Symbol, (tmin_)?NumericQ, (tmax_)?NumericQ}, 
    {restartcond_, restartfn_}, (opts___)?OptionQ] := 
   Module[{initvals, soln, subsoln, diffeqns, tprev, T, k, endvals,
     mintime = MinRecoveryTime /. {opts} /. Options[RestartingNDSolve], 
     ttol = .5/10^TimePrecisionGoal /. {opts} /. Options[RestartingNDSolve], 
     msf = MaxStepFraction /. {opts} /. Options[RestartingNDSolve], 
     report = TrueQ[Reporting /. {opts} /. Options[RestartingNDSolve]], 
     t0 = sys /. {___, _[(s_)?NumericQ] == _ | _ == _[(s_)?NumericQ], ___} -> s}, 
     initvals = vars /. t -> t0 /. First[Solve[sys, vars /. t -> t0]]; 
     diffeqns = DeleteCases[sys, _[t0] == _?NumericQ | _?NumericQ == _[t0]]; 
     If[checkinput[vars, initvals, t, t0, tmin, {restartcond, restartfn}] === $Failed, Return[$Failed]];
     
     k = 0; T[0] = t0;
     While[T[k] < tmax, k++;
       If[report && T[k-1] > t0, Print[StringJoin["Restarting at time t = ", ToString[T[k-1]]]]];

       {soln[k], endvals, tprev, T[k]} = advanceSoln[diffeqns, initvals, vars, {t, T[k-1], tmax}, restartcond, mintime, opts];
       If[!NumericQ[T[k]], Return[$Failed]];
       If[tprev <= T[k-1] + mintime <= T[k], Message[RestartingNDSolve::minrec]];
       If[T[k] < tmax, subsoln = soln[k]; 
         While[(T[k] - tprev)/T[k] > ttol/msf, 
           initvals = vars /. subsoln /. t -> tprev; 
           {subsoln, endvals, tprev, T[k]} = advanceSoln[diffeqns, initvals, vars, {t, tprev, T[k]}, restartcond, 0, opts];
           If[!NumericQ[T[k]],Return[$Failed]];
           ] ];
       If[T[k] < tmax, initvals = restartfn@((vars/.t->T[k])/.endvals)] ;
        ];
     wrapItUp[soln, vars, t, T, k]
     ];
        
RestartingNDSolve[sys:{_Equal..}, vars:{_Symbol..}, {t_Symbol, (t0_)?NumericQ, (tmax_)?NumericQ}, 
    {rc_, rf_}, (opts___)?OptionQ] := RestartingNDSolve[sys, (#1[t] & ) /@ vars, {t, t0, tmax}, {rc, rf}, opts];
     
RestartingNDSolve[sys:{_Equal, _Equal}, (x_)[t_] | (x_Symbol), {t_Symbol, (t0_)?NumericQ, 
     (tmax_)?NumericQ}, {rc_, rf_}, (opts___)?OptionQ] := RestartingNDSolve[sys, {x[t]}, {t, t0, tmax}, {rc, rf}, opts];
     
checkinput[vars_, initvals, t_, t0_, tmin_, {restartcond_, restartfn_}] := 
   Module[{r}, If[t0 != tmin, Message[RestartingNDSolve::badinit]; Return[$Failed]]; 
     If[ !MemberQ[{True, False}, restartcond /. Thread[vars -> initvals]], 
      Message[RestartingNDSolve::badcond]; Return[$Failed]]; 
     r = If[Length[initvals] == 1, {restartfn[First[initvals]]}, restartfn[initvals]]; 
     If[ !MatchQ[r, {_?NumericQ..}] || Length[r] =!= Length[initvals], 
      Message[RestartingNDSolve::badf]; Return[$Failed]]]; 

advanceSoln[diffeqns_, initvals_, vars_, {t_, t1_, t2_}, restartcond_, mintime_, opts___] := 
   Module[{state}, state = First[NDSolve`ProcessEquations[Join[diffeqns, 
         Thread[(vars /. t -> t1) == initvals]], vars, {t, t1, t2}, 
        StoppingTest -> restartcond && t >= mintime + t1, FilterOptions[NDSolve, opts, Options[RestartingNDSolve]]]];
        NDSolve`Iterate[state, t2]; 
     {NDSolve`ProcessSolutions[state], NDSolve`ProcessSolutions[state, "Forward"],
     (state@"CurrentTime"["Forward"]-2*state@"TimeStep"["Forward"]), state@"CurrentTime"["Forward"]}];
       
wrapItUp[soln_, vars_, t_, T_, k_] := (Which @@ #1 &) /@ 
   (Flatten[Transpose[{Append[Table[T[i-1]<=t< T[i], {i,1,k-1}], T[k-1]<=t<=T[k]], #1}]] &) /@ Transpose[vars /. Array[soln, k]];
   
End[]; 
Protect["RNDS`RestartingNDSolve`*"]; 
EndPackage[]; 